import asyncio
import json
import logging
import os
import random
import signal
import time
from typing import Optional, List, Tuple
from aiges.dto import Response, ResponseData, DataListNode, DataListCls, init_rq
from transformers import (AutoModelForCausalLM, AutoTokenizer,
                          PreTrainedTokenizerBase)
"""personal param"""
model_name = 'chinese_llama_alpaca_2'
pretrain_model_path = '/mnt/atpdata/tmp_copy_resource/chinese_llama_alpaca_2'
token_path = '/mnt/atpdata/tmp_copy_resource/chinese_llama_alpaca_2'
lora_list = {
    "1":"/mnt/atpdata/tmp_copy_resource/finetuned_models/alpaca2_couplets_5e/adapter.zip",
    "2":"/mnt/atpdata/tmp_copy_resource/finetuned_models/alpaca2_faq_15e/adapter.zip",
    "3":"/mnt/atpdata/tmp_copy_resource/finetuned_models/alpaca2_wether_50e/adapter.zip",
}

# model_name = 'llama2_7b'
# pretrain_model_path = '/home/tmp_copy_resource/llama2_7b'
# token_path = '/home/tmp_copy_resource/llama2_7b'
# lora_list = {
#     "1":"/home/tmp_copy_resource/finetuned_models/llama2_muying_200e/adapter.zip",
#     "2":"/home/tmp_copy_resource/finetuned_models/llama2_wenan_100e/adapter.zip",
#     "3":"/home/tmp_copy_resource/finetuned_models/llama2_pinglun_30e/adapter.zip",
# }

# model_name = 'llama3_cn_8b_instruct'
# pretrain_model_path = '/mnt/atpdata/tmp_copy_resource/' + model_name
# token_path = pretrain_model_path
# lora_list = {}

model_name = 'internlm2_math_plus_1_8b'
pretrain_model_path = '/workspace/LLamFile/ModelFile/internlm/' + model_name
token_path = pretrain_model_path
lora_list = {}

# 设置为'0'， 代表基座推理
# 设置为'0'之外的任意数字，代表加载lora
kBasePatchID = '0'
select_lora = "0" # '2'

# ShareGPT_V3_unfiltered_cleaned_split 为vllm官方性能测试集
# 下载地址：https://huggingface.co/datasets/anon8231489123/ShareGPT_Vicuna_unfiltered/resolve/main/ShareGPT_V3_unfiltered_cleaned_split.json
benchmark_dataset_path = "/data1/cgzhang6/ShareGPT_Vicuna_unfiltered/ShareGPT_V3_unfiltered_cleaned_split.json"
#benchmark_dataset_path = "/home/tmp_copy_resource/ailabsdk_dataset/nlp/alpaca/product_description_generation/train.json"
# benchmark_dataset_path = "/data1/cgzhang6/muying/train.jsonl"

stream = 1
session_id = 1

finetune_type = 'lora'
infer_type = 'base'
""""""


result_q = init_rq()
import sys
print(sys.path)

wrapper = None
# 创建Logger对象
testlogger = logging.getLogger('my_logger')
# "Write a function that computes the set of sums of all contiguous sublists of a given list in golang."
def get_messages(text):
    message = {
        "messages": [
            {"role": 'user', "content": text}
        ]
    }
    import json
    return json.dumps(message)

def Init():
    import os
    os.environ['PRETRAINED_MODEL_NAME'] = model_name
    os.environ['FULL_MODEL_PATH'] = pretrain_model_path
    os.environ['TOKENIZER_PATH'] = token_path
    os.environ['FINETUNE_TYPE'] = finetune_type
    os.environ['INFER_TYPE'] = infer_type
    import importlib
    module_path = f'ailab.utils.wrapper'
    wrapper_module = importlib.import_module(module_path)
    Wrapper = getattr(wrapper_module, 'Wrapper')
    global wrapper
    wrapper = Wrapper()
    config = {}
    if stream:
        config['common.lic'] = 1
    wrapper.wrapperInit(config)

def Uninit():
    wrapper.wrapperFini()

def LoadRes(patch_id, patch_zip):
    if patch_id == kBasePatchID:
        return
    zip_file_path = patch_zip
    with open(zip_file_path, 'rb') as zip_file:
        # 读取压缩包的二进制数据
        zip_data = zip_file.read()
        # 计算数据长度
        zip_data_length = len(zip_data)

    list_node = DataListNode()
    list_node.key = str(patch_id)
    list_node.data = zip_data
    list_node.len = zip_data_length

    req_data = DataListCls()
    req_data.list.append(list_node)
    wrapper.wrapperLoadRes(req_data, patch_id)

def Once(patch_id, text):
    http_node = DataListNode()
    http_node.key = 'messages'
    text_data = get_messages(text).encode('utf-8')
    http_node.data = text_data
    http_data = DataListCls()
    http_data.list.append(http_node)
    wrapper.wrapperOnceExec({"patch_id":patch_id}, http_data)

def OnceAsync(patch_id, text, sid: str):
    http_node = DataListNode()
    http_node.key = 'messages'
    text_data = get_messages(text).encode('utf-8')
    http_node.data = text_data
    http_data = DataListCls()
    http_data.list.append(http_node)
    wrapper.wrapperOnceExecAsync({"patch_id":patch_id}, http_data, sid)

def UnloadRes(patch_id):
    if patch_id == kBasePatchID:
        return
    wrapper.wrapperUnloadRes(patch_id)

def StreamCreate(patch_id:str, max_tokens = 2048, sid: str = '1234', usertag:str = "9527"):
    params = {
        "patch_id": patch_id,
        "temperature": '0.5',
        "top_k": '4',
        "max_tokens": str(max_tokens)
    }
    s = wrapper.wrapperCreate(params, sid, 0, usertag)
    return s.handle

def StreamWrite(handle,text):
    http_node = DataListNode()
    http_node.key = 'messages'
    text_data = get_messages(text).encode('utf-8')
    http_node.data = text_data
    http_data = DataListCls()
    http_data.list.append(http_node)

    wrapper.wrapperWrite(handle,http_data)

def StreamDestroy(handle):
    wrapper.wrapperDestroy(handle)


def sample_requests(
    dataset_path: str,
    num_requests: int,
    tokenizer: PreTrainedTokenizerBase,
    fixed_output_len: Optional[int] = None,
) -> List[Tuple[str, int, int]]:
    if fixed_output_len is not None and fixed_output_len < 4:
        raise ValueError("output_len too small")

    # Load the dataset.
    with open(dataset_path) as f:
        dataset = json.load(f)
    # Filter out the conversations with less than 2 turns.
    dataset = [data for data in dataset if len(data["conversations"]) >= 2]
    # Only keep the first two turns of each conversation.
    dataset = [(data["conversations"][0]["value"],
                data["conversations"][1]["value"]) for data in dataset]

    # Tokenize the prompts and completions.
    prompts = [prompt for prompt, _ in dataset]
    prompt_token_ids = tokenizer(prompts).input_ids
    completions = [completion for _, completion in dataset]
    completion_token_ids = tokenizer(completions).input_ids
    tokenized_dataset = []
    for i in range(len(dataset)):
        output_len = len(completion_token_ids[i])
        if fixed_output_len is not None:
            output_len = fixed_output_len
        tokenized_dataset.append((prompts[i], prompt_token_ids[i], output_len))

    # Filter out too long sequences.
    filtered_dataset: List[Tuple[str, int, int]] = []
    for prompt, prompt_token_ids, output_len in tokenized_dataset:
        prompt_len = len(prompt_token_ids)
        if prompt_len < 4 or output_len < 4:
            # Prune too short sequences.
            continue
        if prompt_len > 1024 or prompt_len + output_len > 2048:
            # Prune too long sequences.
            continue
        filtered_dataset.append((prompt, prompt_len, output_len))

    # Sample the requests.
    sampled_requests = random.sample(filtered_dataset, num_requests)
    return sampled_requests

def sample_requests_alpaca_dataset(
    dataset_path: str,
    num_requests: int,
    tokenizer: PreTrainedTokenizerBase,
    fixed_output_len: Optional[int] = None,
) -> List[Tuple[str, int, int]]:
    if fixed_output_len is not None and fixed_output_len < 4:
        raise ValueError("output_len too small")

    from datasets import load_dataset
    hf_dataset = load_dataset('json', data_files=dataset_path)
    dataset = list(hf_dataset['train'])
    dataset = [(data['instruction'] + '\n' + data['input'],
                data["output"]) for data in dataset]
    prompts = [prompt for prompt, _ in dataset]
    prompt_token_ids = tokenizer(prompts).input_ids
    completions = [completion for _, completion in dataset]
    completion_token_ids = tokenizer(completions).input_ids
    tokenized_dataset = []
    for i in range(len(dataset)):
        output_len = len(completion_token_ids[i])
        if fixed_output_len is not None:
            output_len = fixed_output_len
        tokenized_dataset.append((prompts[i], prompt_token_ids[i], output_len))

    # Filter out too long sequences.
    filtered_dataset: List[Tuple[str, int, int]] = []
    for prompt, prompt_token_ids, output_len in tokenized_dataset:
        prompt_len = len(prompt_token_ids)
        if prompt_len < 4 or output_len < 4:
            # Prune too short sequences.
            continue
        if prompt_len > 1024 or prompt_len + output_len > 2048:
            # Prune too long sequences.
            continue
        filtered_dataset.append((prompt, prompt_len, output_len))

    # Sample the requests.
    sampled_requests = random.sample(filtered_dataset, num_requests)
    return sampled_requests

import functools
def time_cost(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        elapsed_time = end_time - start_time
        print(f"函数 {func.__name__} 的耗时为: {elapsed_time} 秒")
        return result
    return wrapper


def meminfo():
    import psutil
    # 获取当前进程ID
    pid = psutil.Process().pid

    # 获取当前进程的内存信息
    process = psutil.Process(pid)
    memory_info = process.memory_info()

    # 获取RSS内存占用（以字节为单位）
    rss_memory = memory_info.rss

    # 转换为人类可读的格式
    rss_memory_readable = psutil._common.bytes2human(rss_memory)
    cpu_percent = psutil.cpu_percent()
    memory_percent = psutil.virtual_memory().percent

    testlogger.info(f"CPU utilization: {cpu_percent}%")
    testlogger.info(f"Memory utilization: {memory_percent}%")
    testlogger.info(f"memory val: {rss_memory_readable}")
    testlogger.info("=========================================\n")

def benchmark():
    num_prompts = 1000
    tokenizer = AutoTokenizer.from_pretrained(token_path, trust_remote_code=True)
    if 'ShareGPT_V3_unfiltered' in benchmark_dataset_path:
        requests = sample_requests(benchmark_dataset_path, num_prompts, tokenizer)
    else:
        requests = sample_requests_alpaca_dataset(benchmark_dataset_path, num_prompts, tokenizer)

    start = time.perf_counter()
    sed_id = session_id
    print(f"vllm benchmark start, total request {len(requests)}")
    # 压测时看内存泄露打开
    # while True:
    for prompt, _, output_len in requests:
        patch_id = select_lora
        handle = StreamCreate(patch_id, output_len, str(sed_id), str(sed_id))
        sed_id += 1
        StreamWrite(handle, prompt)
    DataEnd = 2  # 尾数据
    result_num = 0
    total_completion_tokens = 0
    total_prompt_tokens = 0
    total_empty_request = 0
    while True:
        item:Response = result_q.get()
        last_frame = False
        for node in item.list:
            if node.status == DataEnd:
                if node.key == "usage":
                    resp = json.loads(node.data)
                    completion_tokens = resp['completion_tokens']
                    total_completion_tokens += completion_tokens
                    if completion_tokens < 2:
                        total_empty_request += 1
                    prompt_tokens = resp['prompt_tokens']
                    total_prompt_tokens += prompt_tokens
                    #print(f"cur:{prompt_tokens}-{completion_tokens}, total:{total_prompt_tokens}-{total_completion_tokens}")
                last_frame = True
        result_q.task_done()
        if last_frame:
            result_num += 1
            if result_num%100 == 0:
                meminfo()
        if result_num == len(requests):
            break

    end = time.perf_counter()
    elapsed_time = end - start

    total_num_tokens = sum(prompt_len + output_len
                           for _, prompt_len, output_len in requests)
    print(f"benchmark end, total prompts: {len(requests)}, vllm Throughput: {len(requests) / elapsed_time:.2f} requests/s, "
          f"vllm:{total_num_tokens / elapsed_time:.2f} tokens/s, "
          f"actual in:{total_prompt_tokens / elapsed_time:.2f} tokens/s, "
          f"actual out:{total_completion_tokens / elapsed_time:.2f} tokens/s, "
          f"actual Throughput:{(len(requests) - total_empty_request) / elapsed_time:.2f} requests/s, "
          f"empty response:{total_empty_request}")
    Uninit()

    sys.exit()

def signal_handler(sig, frame):
    print("start exit progress")
    Uninit()
    sys.exit()

if __name__ == '__main__' :
    # 获取当前脚本的路径
    script_path = os.path.abspath(__file__)
    # 获取当前脚本所在的目录
    script_directory = os.path.dirname(script_path)
    # 构建 input.txt 文件的路径
    input_file_path = os.path.join(script_directory, 'input.txt')
    print(input_file_path)


    # 配置日志文件的路径和格式
    file_handler = logging.FileHandler('btest1.log')
    formatter = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
    file_handler.setFormatter(formatter)

    # 添加文件处理器到Logger对象
    testlogger.addHandler(file_handler)

    # 设置日志级别
    testlogger.setLevel(logging.INFO)

    Init()
    # 注册信号处理函数
    signal.signal(signal.SIGINT, signal_handler)


    for key, value in lora_list.items():
        LoadRes(key, value)

    # 不同lora直接的并发测试
    # input_str = "what is nlp"
    # handle = StreamCreate("1", usertag="123")
    # StreamWrite(handle, input_str)
    # #time.sleep(0.5)
    # handle = StreamCreate(kBasePatchID, usertag="124")
    # StreamWrite(handle, input_str)
    # handle = StreamCreate("2", usertag="125")
    # StreamWrite(handle, input_str)

    # 性能测试
    #benchmark()

    # once测试
    # Once(kBasePatchID, 'What is natural language processing?')
    # Once(select_lora, 'What is natural language processing?')

    # once异步接口测试
    # OnceAsync(kBasePatchID, 'What is natural language processing?', session_id)
    # OnceAsync(select_lora, 'What is natural language processing?', str(session_id+1))

    while True:
        patch_id = select_lora
        max_tokens = 2048
        infer_type = "base" if patch_id == kBasePatchID else f"patch_id:{patch_id}"
        input_str = input(f"==>1:read prompt from input.txt q/exit:exit({model_name}+{infer_type}):")
        if input_str == 'q' or input_str == 'exit':
            Uninit()
            sys.exit()
        else:
            try:
                input_num = int(input_str)
            except ValueError:
                continue
            with open(input_file_path, 'r', encoding='utf-8') as file:
                input_str = file.read()
            if input_num != 1:
                max_tokens = input_num
        handle = StreamCreate(patch_id, max_tokens=max_tokens)
        StreamWrite(handle, input_str)
        # 异步返回，不能在此处销毁，否则还没推理完就销毁了
        #StreamDestroy(handle)

    #UnloadRes('1')


