import cv2
import numpy as np
from ailab.log import logger

def get_current_time():
    from datetime import datetime
    from datetime import timezone,timedelta
    current_time =  datetime.now(timezone(timedelta(hours=8)))
    utc = current_time.strftime("%Y-%m-%d %H:%M:%S:%f")[:-3]
    return utc

def get_current_timestamp_13():
    from datetime import datetime
    from datetime import timezone,timedelta
    current_time =  datetime.now(timezone(timedelta(hours=8)))
    timestamp = current_time.timestamp() * 1000  # 转换为毫秒
    timestamp_13 = int(timestamp)  # 转换为整数并截取前 13 位
    return timestamp_13

def add_watermark_on_image(image:np.ndarray, watermark_image:np.ndarray, 
                           image_height, image_width, position, wmk_scale, transparency):
    wmk_x, wmk_y, wmk_w, wmk_h = position
    wmk_w = int(wmk_w * wmk_scale)
    wmk_h = int(wmk_h * wmk_scale)
    wmk_x = image_width - wmk_w
    wmk_y = image_height - wmk_h
    wmk_ratio = (wmk_w * wmk_h) / (image_width * image_height)
    wmk_ref_ratio = 0.03
    if wmk_ratio < wmk_ref_ratio:
        print(f"watermark area less than 3%:{wmk_ratio}")

    if wmk_x == -1 or wmk_y == -1:
        wmk_right_space = wmk_w // 2
        wmk_bottom_space = wmk_h // 2
        wmk_x = max(0, image_width - wmk_right_space - wmk_w)
        wmk_y = max(0, image_height - wmk_bottom_space - wmk_h)

    if wmk_x + wmk_w > image_width:
        if wmk_w > image_width:
            wmk_x = 0
            wmk_w = image_width
        else:
            wmk_x = image_width - wmk_w

    rect = (wmk_x, wmk_y, wmk_w, wmk_h)
    image_region = image[wmk_y:wmk_y+wmk_h, wmk_x:wmk_x+wmk_w].astype(np.float32)
    
    watermark_image_32f = watermark_image.astype(np.float32)
    watermark_image_32f_resized = cv2.resize(watermark_image_32f, (wmk_w, wmk_h))
    watermark_image_channels_rgb = watermark_image_32f_resized[:, :, :3]
    
    if watermark_image.shape[2] == 3:
        alpha_array = np.ones((wmk_h, wmk_w, 3), dtype=np.float32)
    else:
        alpha = watermark_image_32f_resized[:, :, 3]
        alpha_array = cv2.cvtColor(alpha, cv2.COLOR_GRAY2RGB) / 255.0

    fuse_matrix = alpha_array * (1 - transparency)
    ones_array = np.ones((wmk_h, wmk_w, 3), dtype=np.float32)
    
    tmp1 = ones_array - fuse_matrix
    tmp2 = tmp1 * image_region
    tmp3 = fuse_matrix * watermark_image_channels_rgb
    
    fuse_region_array = tmp2 + tmp3
    fuse_region_array_8u = fuse_region_array.astype(np.uint8)
    
    for i in range(fuse_region_array_8u.shape[0]):
        for j in range(fuse_region_array_8u.shape[1]):
            image[wmk_y + i, wmk_x + j] = fuse_region_array_8u[i, j]
    return image

def add_water_on_corners(image: np.ndarray, uid: str, sid: str):
    exif = "{\"company\":\"iflytek\",\"uid\":\""
    if len(uid) != 32:
        uid = "00000000000000000000000000000000"
    if len(sid) != 32:
        sid = "00000000000000000000000000000000"
    timestamp = str(get_current_timestamp_13())
    exif += uid + "\",\"sid\":\"" + sid + "\",\"timestamp\":\"" + timestamp + "\"}"
    cords = [
        [0, 0, image.shape[1]//2, image.shape[0]//2],
        [image.shape[1]//2, 0, image.shape[1]//2, image.shape[0]//2],
        [0, image.shape[0]//2, image.shape[1]//2, image.shape[0]//2],
        [image.shape[1]//2, image.shape[0]//2, image.shape[1]//2, image.shape[0]//2]
    ]

    region = 35
    w, h = image.shape[1]//2, image.shape[0]//2

    for i in range(4):
        x1, y1, x2, y2 = cords[i]
        image_region = image[y1:y1+y2, x1:x1+x2]

        watermark_bin = ''
        for c in exif:
            bits = format(ord(c), '08b')
            watermark_bin += bits
        
        sx, sy = w//2, h//2

        if sx + region > w:
            sx = max(w - region, 0)
        if sy + region > h:
            sy = max(h - region, 0)
        
        ind = 0
        for c in watermark_bin:
            x = sx + ind % region
            y = sy + ind // region
            ori_pixel = image[y, x, 2]
            re_pixel = (ori_pixel & 0xFE) | (int(c) - 0)
            image[y, x, 2] = re_pixel
            ind += 1

        for r in range(image_region.shape[0]):
            for c in range(image_region.shape[1]):
                image[r + y1, c + x1] = image_region[r, c]
        return image

import io
from PIL import Image
from PIL.PngImagePlugin import PngInfo
def watermark_header(image: np.ndarray, sid):
    # Create an array to store PNG image bytes
    width, height = image.shape[:2]
    metadata_string = f'AIGC:{{"ServiceProvider":"iflytek","Time":"{get_current_time()}","ContentID":"{sid}"}}'
    # 检查图像通道并转换为RGB
    if image.shape[2] == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    # 转换OpenCV图像到Pillow图像
    pil_image = Image.fromarray(image)

    # 创建PngInfo对象，用于存储元数据
    png_info = PngInfo()
    png_info.add_text("exif", metadata_string)
    # 创建内存文件
    img_byte_arr = io.BytesIO()
    # 保存图像到内存中的PNG文件，包括元数据
    pil_image.save(img_byte_arr, format='PNG', pnginfo=png_info)
    img_byte_arr = img_byte_arr.getvalue()  # 获取二进制数据
    return img_byte_arr

def jpg_custom_header(image: np.ndarray, sid):
    # 读取JPEG数据
    metadata_string = f'AIGC:{{"ServiceProvider":"iflytek","Time":"{get_current_time()}","ContentID":"{sid}"}}'
    # 转换OpenCV图像到Pillow图像
    if image.shape[2] == 3:
        image = cv2.cvtColor(image, cv2.COLOR_BGR2RGB)
    img = Image.fromarray(image)

    import piexif
    exif_info = {
        "Exif": {
            piexif.ExifIFD.UserComment: metadata_string.encode()
        }
    }
    # 将EXIF字典转换为字节
    exif_bytes = piexif.dump(exif_info)

    # 将图像保存为JPEG，并返回数据
    output_data = io.BytesIO()
    img.save(output_data, format='JPEG',exif=exif_bytes)
    return output_data.getvalue()
    
def png_header_test(image):
    # 使用函数添加水印
    sid = '00000ggg8000000000g0080000000000'  # 假定的ContentID -32位
    image_header = jpg_custom_header(image, sid)
    return image_header

def watermask_corner_test(image):
    # 使用示例
    uid = "00000000000000000000000000000000"
    sid = '00000ggg8000000000g0080000000000'
    image_inside = add_water_on_corners(image, uid, sid)
    # 保存图像
    return image_inside

def calc_wmk_scale(image_width,image_height,wmk_width,wmk_height):
    return (image_width * image_height) * 0.02 / (wmk_width * wmk_height)

def upside_water_test(image):
    import cv2
    import numpy as np
    # 加载图像和水印图像 
    watermark_image = cv2.imread('/home/png/watermask.png', cv2.IMREAD_UNCHANGED)
    # 确保水印图像有4个通道（包括alpha通道）
    if watermark_image.shape[2] == 3:
        watermark_image = cv2.cvtColor(watermark_image, cv2.COLOR_BGR2BGRA)
    # 设置其他参数
    image_height, image_width = image.shape[:2]
    wmk_height, wmk_width = watermark_image.shape[:2]
    position = (image_width - wmk_width, image_height - wmk_height, wmk_width, wmk_height)  # 示例位置，这里是[x, y, width, height]
    wmk_scale = calc_wmk_scale(image_width,image_height,wmk_width,wmk_height)
    transparency = 0.7
    # 调用函数
    result_image = add_watermark_on_image(image, watermark_image, image_height, image_width, position, wmk_scale, transparency)
    # 保存结果图像
    return result_image

if __name__ == '__main__' :
    image = cv2.imread('/home/png/cat.jpg') 
    image = upside_water_test(image)
    sid = '00000ggg8000000000g0080000000000'  # 假定的ContentID -32位
    image_data = jpg_custom_header(image, sid)
    
    """
    image = watermask_corner_test(image)
    image_data = png_header_test(image)
    """

    byte_stream = io.BytesIO(image_data)
    from PIL import Image
    image = Image.open(byte_stream)

    exif_data = ""
    for k,v in image.info.items():
        if k == 'exif':
            exif_data = v
            print(f'exif_data {exif_data}')
            break

    """
    exif_data = exif_data.decode()
    key_list = ["AIGC","ServiceProvider","iflytek","Time","ContentID"]
    for k in key_list:
        if k not in exif_data:
            print(f'k {k} failed exif_data')
    """
    # 保存图像到指定目录
    with open('result.jpg', 'wb') as file:
        file.write(image_data)
    